@@ -1,5 +1,5 @@
 /* Log file output.
-   Copyright (C) 2003, 2005 Free Software Foundation, Inc.
+   Copyright (C) 2003, 2005, 2009 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU Library General Public License as published
@@ -33,24 +33,27 @@
 # include "lock.h"
 #endif
 
+/* Separator between msgctxt and msgid in .mo files.  */
+#define MSGCTXT_SEPARATOR '\004'  /* EOT */
+
 /* Print an ASCII string with quotes and escape sequences where needed.  */
 static void
-print_escaped (FILE *stream, const char *str)
+print_escaped (FILE *stream, const char *str, const char *str_end)
 {
   putc ('"', stream);
-  for (; *str != '\0'; str++)
+  for (; str != str_end; str++)
     if (*str == '\n')
       {
-	fputs ("\\n\"", stream);
-	if (str[1] == '\0')
-	  return;
-	fputs ("\n\"", stream);
+        fputs ("\\n\"", stream);
+        if (str + 1 == str_end)
+          return;
+        fputs ("\n\"", stream);
       }
     else
       {
-	if (*str == '"' || *str == '\\')
-	  putc ('\\', stream);
-	putc (*str, stream);
+        if (*str == '"' || *str == '\\')
+          putc ('\\', stream);
+        putc (*str, stream);
       }
   putc ('"', stream);
 }
@@ -61,9 +64,10 @@ __libc_lock_define_initialized (static, lock)
 
 static inline void
 _nl_log_untranslated_locked (const char *logfilename, const char *domainname,
-			     const char *msgid1, const char *msgid2, int plural)
+                             const char *msgid1, const char *msgid2, int plural)
 {
   FILE *logfile;
+  const char *separator;
   size_t len;
 
   /* Can we reuse the last opened logfile?  */
@@ -71,35 +75,43 @@ _nl_log_untranslated_locked (const char *logfilename, const char *domainname,
     {
       /* Close the last used logfile.  */
       if (last_logfilename != NULL)
-	{
-	  if (last_logfile != NULL)
-	    {
-	      fclose (last_logfile);
-	      last_logfile = NULL;
-	    }
-	  free (last_logfilename);
-	  last_logfilename = NULL;
-	}
+        {
+          if (last_logfile != NULL)
+            {
+              fclose (last_logfile);
+              last_logfile = NULL;
+            }
+          free (last_logfilename);
+          last_logfilename = NULL;
+        }
       /* Open the logfile.  */
       len = strlen (logfilename) + 1;
       last_logfilename = (char *) malloc (len);
       if (last_logfilename == NULL)
-	return;
+        return;
       strlcpy (last_logfilename, logfilename, len);
       last_logfile = fopen (logfilename, "a");
       if (last_logfile == NULL)
-	return;
+        return;
     }
   logfile = last_logfile;
 
   fprintf (logfile, "domain ");
-  print_escaped (logfile, domainname);
+  print_escaped (logfile, domainname, domainname + strlen (domainname));
+  separator = strchr (msgid1, MSGCTXT_SEPARATOR);
+  if (separator != NULL)
+    {
+      /* The part before the MSGCTXT_SEPARATOR is the msgctxt.  */
+      fprintf (logfile, "\nmsgctxt ");
+      print_escaped (logfile, msgid1, separator);
+      msgid1 = separator + 1;
+    }
   fprintf (logfile, "\nmsgid ");
-  print_escaped (logfile, msgid1);
+  print_escaped (logfile, msgid1, msgid1 + strlen (msgid1));
   if (plural)
     {
       fprintf (logfile, "\nmsgid_plural ");
-      print_escaped (logfile, msgid2);
+      print_escaped (logfile, msgid2, msgid2 + strlen (msgid2));
       fprintf (logfile, "\nmsgstr[0] \"\"\n");
     }
   else
@@ -110,7 +122,7 @@ _nl_log_untranslated_locked (const char *logfilename, const char *domainname,
 /* Add to the log file an entry denoting a failed translation.  */
 void
 _nl_log_untranslated (const char *logfilename, const char *domainname,
-		      const char *msgid1, const char *msgid2, int plural)
+                      const char *msgid1, const char *msgid2, int plural)
 {
   __libc_lock_lock (lock);
   _nl_log_untranslated_locked (logfilename, domainname, msgid1, msgid2, plural);
